"""Reference device for testing the individual report format
"""

from iotile.core.hw.virtual.virtualdevice import VirtualIOTileDevice, rpc, RPCInvalidIDError, RPCNotFoundError, TileNotFoundError
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IndividualReadingReport, IOTileReading, SignedListReport

class ReportTestDevice(VirtualIOTileDevice):
    """Mock IOTileDevice that creates a sequence of reports upon connection

    This device can be considered a reference implementation of the individual
    reading and signed list report formats.

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            Supported args are:
                iotile_id (int): The UUID used for this device (default: 1)
                num_readings (int): The number of readings to be generated every
                    time the streaming interface is opened. (default: 100)
                starting_timestamp (int): The first timestamp that should be generated
                    for the first Reading sent.  Timestamps are sequential for each
                    reading, so x, x+1, etc.
                    Default: 0
                starting_id (int): The first reading id that should be generated
                    for the first Reading sent.  Reading IDs will be sequential after
                    starting_id so x, x+1, x+2.
                    Default: 1
                reading_generator (string): The method for generating readings. 
                    Options are: sequential, random (default: sequential)
                    random will generate random values between the reading_min and
                    reading_max keys (default: 0, 100)
                    sequential will sequentially generate reading values starting
                    at reading_start (default: 0)
                format (string): The report format to package in (either individual
                or signed_list).  (default: individual)
                report_length (int): The maximum number of readings per report
                    (default: 10).  The only applies to report formats that can contain
                    multiple readings
                signing_method (int): The signature type to be applied to signed messages
                    Common values would be 0 for a hash only signature with no origin 
                    verification functionality or 1 for signing with a user set key.  
                    (default: 0)
                stream_id (int): The stream that this reading should be sent out
    """

    def __init__(self, args):
        iotile_id = args.get('iotile_id', 1)

        if isinstance(iotile_id, basestring) or isinstance(iotile_id, unicode):
            iotile_id = int(iotile_id, 16)

        generator = args.get('reading_generator', 'sequential')            
        self.num_readings = args.get('num_readings', 100)

        stream_string = args.get('stream_id', '5001')
        self.stream_id = int(stream_string, 16)

        self.signing_method = args.get('signing_method', 0)
        self.start_timestamp = args.get('starting_timestamp', 0)
        self.start_id = args.get('starting_id', 1)
        
        #Now pull in args for the generator
        if generator == 'sequential':
            self.reading_start = args.get('reading_start', 0)
            self.generator = self._generate_sequential
        elif generator == 'random':
            self.reading_min = args.get('reading_min', 0)
            self.reading_max = args.get('reading_max', 100)
            self.generator = self._generate_random
        else:
            raise ArgumentError("Unknown reading generator mechanism", reading_generator=generator, known_generators=['sequential', 'random'])

        self.format = args.get('format', 'individual')
        if self.format not in ['individual', 'signed_list']:
            raise ArgumentError("Unknown report format for generator", format=self.format, known_formats=['individual', 'signed_list'])

        self.report_length = args.get('report_length', 10)

        if self.report_length == 0 and self.format != 'individual' and self.num_readings != 0:
            raise ArgumentError("You cannot have a report length of 0 and more than 0 readings because that would be an infinite loop")

        super(ReportTestDevice, self).__init__(iotile_id, 'Simple')

    @rpc(8, 0x0004, "", "H6sBBBB")
    def controller_name(self):
        """Return the name of the controller as a 6 byte string
        """

        status = (1 << 1) | (1 << 0) #Configured and running
        
        return [0xFFFF, self.name, 1, 0, 0, status]

    def _generate_sequential(self):
        readings = []
        
        for i in xrange(self.reading_start, self.num_readings+self.reading_start):
            reading = IOTileReading(i-self.reading_start+self.start_timestamp, self.stream_id, i, reading_id=i-self.reading_start+self.start_id)
            readings.append(reading)

        return readings

    def _generate_random(self):
        return []

    def open_streaming_interface(self):
        """Called when someone opens a streaming interface to the device

        Returns:
            list: A list of IOTileReport objects that should be sent out
                the streaming interface.
        """

        readings = self.generator()

        reports = []

        if self.format == 'individual':
            reports = [IndividualReadingReport.FromReadings(self.iotile_id, [reading]) for reading in readings]
        elif self.format == 'signed_list':
            if self.report_length == 0:
                return [SignedListReport.FromReadings(self.iotile_id, [], signature_method=self.signing_method)]

            for i in xrange(0, len(readings), self.report_length):
                chunk = readings[i:i+self.report_length]
                report = SignedListReport.FromReadings(self.iotile_id, chunk, signature_method=self.signing_method)
                reports.append(report)

        return reports



import unittest
import pytest
import os.path
import os
import struct
import logging
import sys
from iotile.mock.mock_iotile import MockIOTileDevice
from iotile.mock.mock_ble import MockBLEDevice

class TestMockBLEBasic(unittest.TestCase):
    def setUp(self):
        self.dev = MockIOTileDevice('TestCN')
        self.ble = MockBLEDevice("AA:BB:CC:DD:EE:FF", self.dev)
        logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

    def tearDown(self):
        pass

    def test_basic_functionality(self):
        """Make sure that things are setup right
        """

        assert self.ble.notifications_enabled(self.ble.TBReceiveHeaderChar) is False
        assert self.ble.notifications_enabled(self.ble.TBReceivePayloadChar) is False
        assert self.ble.notifications_enabled(self.ble.TBStreamingChar) is False

    def test_calling_invalid_rpc(self):
        """Make sure that basically calling an RPC works
        """

        header = struct.pack("<BBBBB", 0, 0, 0xFF, 0xFF, 127)
        
        success, notif = self.ble._handle_write(self.ble.TBSendPayloadChar, "")
        assert success is True
        assert len(notif) == 0

        success, notif = self.ble._handle_write(self.ble.TBSendHeaderChar, header)
        assert success is True
        assert len(notif) == 1

        char_id, resp_header = notif[0]
        assert self.ble.find_uuid(char_id)[0] == self.ble.TBReceiveHeaderChar

        print len(resp_header)
        status, _, resp_len, _ = struct.unpack("<BBBB", resp_header)
        assert status == 0xFF
        assert resp_len == 0

    def test_calling_valid_rpc(self):
        """Make sure that calling a valid RPC with a payload works
        """

        header = struct.pack("<BBBBB", 0, 0, 0x04, 0x00, 8)

        success, notif = self.ble._handle_write(self.ble.TBSendPayloadChar, "")
        assert success is True
        assert len(notif) == 0

        success, notif = self.ble._handle_write(self.ble.TBSendHeaderChar, header)
        print notif
        assert success is True
        assert len(notif) == 2

        char_id, resp_header = notif[0]
        assert self.ble.find_uuid(char_id)[0] == self.ble.TBReceiveHeaderChar

        char_id, resp_payload = notif[1]
        assert self.ble.find_uuid(char_id)[0] == self.ble.TBReceivePayloadChar


from iotile.mock.devices.report_test_device import ReportTestDevice

def test_report_args():
    """Make sure report arguments are properly parsed and used
    """

    args = {
        "iotile_id": "1",
        "reading_start": 100,
        "num_readings": 107,
        "stream_id": "5001",
        "format": "signed_list",
        "report_length": 10,
        "signing_method": 0,
        'starting_timestamp': 10,
        'starting_id': 15
    }

    dev = ReportTestDevice(args)

    assert dev.start_timestamp == 10
    assert dev.start_id == 15

    reports = dev.open_streaming_interface()

    assert len(reports) == 11

    report1 = reports[0]

    assert report1.lowest_id == 15
    assert report1.highest_id == 24 
    assert len(report1.visible_readings) == 10
    
    read = report1.visible_readings[0]
    assert read.stream == 0x5001
    assert read.value == 100
    assert read.raw_time == 10



import argparse
import pkg_resources
import sys
import logging
import json
import os.path
import imp
from iotile.core.exceptions import ArgumentError, IOTileException
from iotile.core.hw.hwmanager import HardwareManager

def main():
    """Run a script that puts an IOTile device into a known state
    """

    list_parser = argparse.ArgumentParser(add_help=False)
    list_parser.add_argument('-l', '--list', action='store_true', help="List all known device preparation scripts and then exit")

    parser = argparse.ArgumentParser(description="Prepare a device or a list of devices into a known state using a script")
    
    parser.add_argument('port', help="The name of the port to use to connect to the device")
    parser.add_argument('script', help="The name of the device preparation script to use (can either be an installed script name or a .py file with extension")
    parser.add_argument('-c', '--config', help="An optional JSON config file with arguments for the script")
    parser.add_argument('-l', '--list', action='store_true', help="List all known scripts and then exit")
    parser.add_argument('--uuid', action='append', default=[], help="Run script on device given by this uuid")
    parser.add_argument('--device', action='append', default=[], help="Run script on device given by this connection string")
    parser.add_argument('--pause', action='store_true', help="Pause and wait for user input after finishing each device")
    parser.add_argument('--max-attempts', type=int, default=1, help="Number of times to attempt the operation (up to a max of 5 times)")
    parser.add_argument('--uuid-range', action='append', default=[], help="Process every device in a range (range should be specified as start-end and is inclusive, e.g ab-cd)")
    args, rest = list_parser.parse_known_args()

    if args.list:
        print("\nInstalled Preparation Scripts:")
        for entry in pkg_resources.iter_entry_points('iotile.device_recipe'):
            print('- {}'.format(entry.name))

        return 0

    args = parser.parse_args()

    config = {}
    if args.config is not None:
        with open(args.config, "rb") as conf_file:
            config = json.load(conf_file)

    script = instantiate_script(args.script)

    success = []

    devices = []
    devices.extend([('uuid', int(x, 16)) for x in args.uuid])
    devices.extend([('conection_string', x) for x in args.device])

    for uuid_range in args.uuid_range:
        start,_,end = uuid_range.partition('-')
        start = int(start, 16)
        end = int(end, 16)

        devices.extend([('uuid', x) for x in xrange(start, end+1)])

    try:
        with HardwareManager(port=args.port) as hw:
            for conntype, dev in devices:
                for i in xrange(0, args.max_attempts):
                    try:
                        print("Configuring device %s identified by %s" % (str(dev), conntype))
                        configure_device(hw, conntype, dev, script, config)
                        break
                    except IOTileException, exc:
                        print("--> Error on try %d: %s" % (i+1, str(exc)))
                        continue

                success.append((conntype, dev))
                if args.pause:
                    raw_input("--> Waiting for <return> before processing next device")
    except KeyboardInterrupt:
        print("Break received, cleanly exiting...")

    print("\n**FINISHED**\n")
    print("Successfully processed %d devices" % len(success))
    for conntype,conn in success:
        print("%s: %s" % (conntype, conn))

    if len(success) != len(devices):
        return 1

    return 0

def import_device_script(script_path):
    """Import a main function from a script file

    script_path must point to a python file ending in .py that contains exactly one
    VirtualIOTileDevice class definitions.  That class is loaded and executed as if it
    were installed.

    Args:
        script_path (string): The path to the script to load

    Returns:
        VirtualIOTileDevice: A subclass of VirtualIOTileDevice that was loaded from script_path
    """

    search_dir, filename = os.path.split(script_path)
    if search_dir == '':
        search_dir = './'

    if filename == '' or not os.path.exists(script_path):
        print("Could not find script to load virtual device, path was %s" % script_path)
        sys.exit(1)

    module_name, ext = os.path.splitext(filename)
    if ext != '.py':
        print("Script did not end with .py")
        sys.exit(1)

    try:
        file = None
        file, pathname, desc = imp.find_module(module_name, [search_dir])
        mod = imp.load_module(module_name, file, pathname, desc)
    finally:
        if file is not None:
            file.close()

    if not hasattr(mod, 'main'):
        print("Script file had no main function containing a device recipe")
        sys.exit(1)

    return getattr(mod, 'main')

def configure_device(hw, conntype, conarg, script, args):
    if conntype == 'uuid':
        hw.connect(conarg)
    else:
        hw.connect_direct(conarg)

    try:
        script(hw, args)
    finally:
        if hw.stream.connected:
            hw.disconnect()

def instantiate_script(device_recipe):
    """Find a device recipe by name and instantiate it

    Args:
        device_recipe (string): The name of the pkg_resources entry point corresponding to
            the device.  It should be in group iotile.device_recipe

    Returns:
        tuple(callable, dict): A callable function with signature callable(HWManager, config_dict) that
            executes the script and a dictionary that must be kept around as long as callable is.
    """

    if device_recipe.endswith('.py'):
        main_func = import_device_script(device_recipe)
        return main_func

    for entry in pkg_resources.iter_entry_points('iotile.device_recipe', name=device_recipe):
        dev = entry.load()
        return dev

    print("Could not find an installed device preparation script with the given name: {}".format(device_recipe))
    sys.exit(1)



import unittest
import pytest
import os.path
import os
import struct
import logging
import sys
from iotile.mock.mock_iotile import MockIOTileDevice
from iotile.mock.mock_adapter import MockDeviceAdapter

class TestMockBLEBasic(unittest.TestCase):
    def setUp(self):
        self.dev = MockIOTileDevice(1, 'TestCN')
        self.adapter = MockDeviceAdapter()
        self.adapter.add_device('test', self.dev)

        logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)

    def test_connect(self):
        res = self.adapter.connect_sync(1, 'test')
        assert res['success'] is True

    def test_connect_nothing(self):
        res = self.adapter.connect_sync(1, 'nothing')
        assert res['success'] is False

    def test_disconnect(self):
        res = self.adapter.connect_sync(1, 'test')
        assert res['success'] is True
        res = self.adapter.disconnect_sync(1)
        assert res['success'] is True

    def test_rpc(self):
        self.adapter.connect_sync(1, 'test')

        res = self.adapter.send_rpc_sync(1, 8, 4, '', 1.0) 
        assert len(res['payload']) == 6
        assert res['payload'] == 'TestCN'

    def test_advertisement(self):
        self.adapter.advertise()



"""Virtual device that acts like a blank tile with just an executive installed
"""

from iotile.core.hw.virtual.virtualdevice import VirtualIOTileDevice, rpc, RPCInvalidIDError, RPCNotFoundError, TileNotFoundError

class NoAppVirtualDevice(VirtualIOTileDevice):
    """Virtual device that acts like a blank tile with just an executive loaded

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            None are supported.
    """

    def __init__(self, args):
        super(NoAppVirtualDevice, self).__init__(1, 'NO APP')

    @rpc(8, 0x0004, "", "H6sBBBB")
    def status(self):
        """Return the name of the controller as a 6 byte string
        """

        status = (1 << 1) | (1 << 0) #Configured and running
        
        return [0xFFFF, self.name, 1, 0, 0, status]



"""A mock DependencyResolver object for testing iotile-build
"""

from iotile.core.dev.iotileobj import IOTile

class MockDependencyResolver(object):
    def __init__(self, args):
        self.known_deps = {IOTile(x).unique_id: IOTile(x) for x in args}

    def resolve(self, depinfo, destdir):
        """Attempt to resolve this dependency using a subclass defined method
        
        Args:
            depinfo (dict): a dictionary returned by IOTile.dependencies describing the dependency
            destdir (string): the directory that the dependency should be copied into

        Returns:
            dict: The function returns a dictionary that has required and optional keys.  The 
                required keys are:
                    found: boolean if this resolver found a matching dependency

                optional keys are:
                    stop:   boolean if this resolver is suggesting that we should stop looking for this
                            dependency. This is useful for making a resolver that stops a resolver chain
                    info:   a dictionary that contains information that this resolver wants to store
                            with the dependency for future reference.
        """

        if depinfo['unique_id'] not in self.known_deps:
            return {'found': False}

        dep = self.known_deps[depinfo['unique_id']]
        if depinfo['required_version'].check(dep.parsed_version):
            self._copy_folder_contents(dep.output_folder, destdir)
            return {'found': True}

        return {'found': False}

    def check(self, depinfo, deptile, depsettings):
        """Check if this dependency is the latest version in a subclass defined way


        Args:
            depinfo (dict): a dictionary returned by IOTile.dependencies describing the dependency
            deptile (IOTile): an IOTile object for the installed dependency
            depsettings (dict): a dictionary that was previously stored with this dependency by resolve

        Returns:
            bool: True meaning the dependency is up-to-date or False if it is not.

        Raises:
            EnvironmentError: if the checking process was not able to assess whether the dependency was 
                up to date or not.
        """

        if depinfo['unique_id'] not in self.known_deps:
            raise EnvironmentError("Could not check dependency's status", unique_id=depinfo['unique_id'])

        dep = self.known_deps[depinfo['unique_id']]
        if not depinfo['required_version'].check(dep.parsed_version):
            return True

        if deptile.parsed_version < dep.parsed_version:
            return False

        return True

    def _copy_folder_contents(self, source, dest):
        import shutil
        shutil.copytree(source, dest)


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.hw.reports.signed_list_format import SignedListReport
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
import unittest
import pytest
import os.path
import os

@pytest.fixture
def simple_hw():
    hw = HardwareManager('virtual:simple')
    yield hw

    hw.disconnect()


@pytest.fixture
def report_hw():
    hw = HardwareManager('virtual:report_test')
    yield hw

    hw.disconnect()

@pytest.fixture
def conf_report_hw():
    conf_file = os.path.join(os.path.dirname(__file__), 'report_zero_length.json')

    if '@' in conf_file or ',' in conf_file or ';' in conf_file:
        pytest.skip('Cannot pass device config because path has [@,;] in it')

    hw = HardwareManager('virtual:report_test@%s' % conf_file)
    yield hw

    hw.disconnect()

def test_config_file(conf_report_hw):
    """Make sure zero length reports work
    """

    conf_report_hw.connect_direct('2')
    conf_report_hw.enable_streaming()

    assert conf_report_hw.count_reports() == 1
    rep1 = [x for x in conf_report_hw.iter_reports()][0]

    assert rep1.verified == 1
    assert len(rep1.visible_readings) == 0

def test_invalid_length_combo():
    """Make sure invalid length combinations throw an exception
    """

    conf_file = os.path.join(os.path.dirname(__file__), 'report_length_invalid.json')

    if '@' in conf_file or ',' in conf_file or ';' in conf_file:
        pytest.skip('Cannot pass device config because path has [@,;] in it')

    with pytest.raises(ArgumentError):
        hw = HardwareManager('virtual:report_test@%s' % conf_file)



"""Basic virtual IOTile device class for testing tracing data from an IOTile device
"""

import binascii
from iotile.core.hw.virtual.virtualdevice import VirtualIOTileDevice, rpc, RPCInvalidIDError, RPCNotFoundError, TileNotFoundError

class TracingTestDevice(VirtualIOTileDevice):
    """Mock IOTileDevice that sends tracing data from the device

    If no arguments are passed, the fixed string 'Hello world, this is tracing data!'
    is sent over the tracing interface.

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            Two mutually exclusive arguments are supported.

            If 'hex_data' is  present, it must be a hex string that is decoded
            into binary and then sent in 20 byte chunks over the tracing
            interface when it is opened.

            If 'ascii_data' is present, it must be an ascii string that is
            sent in 20 byte chunks with no decoding.

            If both keys are specified, ascii_data wins.

    """

    def __init__(self, args):
        super(TracingTestDevice, self).__init__(1, 'Simple')

        data = bytearray("Hello world, this is tracing data!")

        if 'hex_data' in args:
            data = bytearray(binascii.unhexlify(args['hex_data']))

        if 'ascii_data' in args:
            data = bytearray(args['ascii_data'].encode('ascii'))

        # Create chunks of tracing data in 20 byte segments to simulate
        # what typically comes from real devices
        for i in xrange(0, len(data), 20):
            self.traces.append(data[i:i+20])

    @rpc(8, 0x0004, "", "H6sBBBB")
    def controller_name(self):
        """Return the name of the controller as a 6 byte string
        """

        status = (1 << 1) | (1 << 0)  # Configured and running

        return [0xFFFF, self.name, 1, 0, 0, status]



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#Caveats and possible issues
#Mac OS X
# - when using a virtualenv, readline is not properly installed into the virtualenv
#   and cannot be imported.  You need to install it using easy_install as described here
#   http://calvinx.com/tag/readline/

import re

from setuptools import setup, find_packages
import version

setup(
    name="iotile-test",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    description="IOTile Test Infrastructure",
    entry_points={'iotile.virtual_device': ['simple = iotile.mock.devices.simple_virtual_device:SimpleVirtualDevice',
                                            'report_test = iotile.mock.devices.report_test_device:ReportTestDevice',
                                            'no_app = iotile.mock.devices.noapp:NoAppVirtualDevice',
                                            'tracing_test = iotile.mock.devices.tracing_test_device:TracingTestDevice'],
                  'iotile.proxy': ['simple = iotile.mock.devices.simple_virtual_proxy'],
                  'console_scripts': ['prepare_device = iotile.test_scripts.prepare_device:main']},
    author="Arch",
    author_email="info@arch-iot.com",
    url="https://github.com/iotile/coretools/iotilecore",
    keywords=["iotile", "arch", "embedded", "hardware"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTileTest
----------

A python package for testing IOTile based infrastructure including mocks for major portions.  

See https://www.arch-iot.com.
"""
)



import pytest
from iotile.core.hw.hwmanager import HardwareManager
import time

def pytest_addoption(parser):
    parser.addoption('--port', default="bled112", help="Port to use to connect to iotile device")
    parser.addoption('--uuid', action="append", default=[], help="Device UUIDs to run tests on")

def pytest_generate_tests(metafunc):
    if 'port' in metafunc.fixturenames:
        port = metafunc.config.option.port

        metafunc.parametrize("port", [port], scope='session')

    if 'device_id' in metafunc.fixturenames:
        ids = [int(x, 0) for x in metafunc.config.option.uuid]
        metafunc.parametrize("device_id", ids, scope='session')

@pytest.fixture(scope='module')
def device(port, device_id):
    """Return a HardwareManager instance connected to an IOTile Device

    This fixture shares the same device among all tests in the same module
    """
    with HardwareManager(port=port) as hw:
        time.sleep(1)

        hw.connect(device_id)
        hw.enable_streaming()
        yield hw
        hw.disconnect()

@pytest.fixture(scope='function')
def per_test_device(port, device_id):
    """Return a HardwareManager instance connected to an IOTile Device

    This fixture creates and tears down the connection for each test
    """
    with HardwareManager(port=port) as hw:
        time.sleep(1)

        hw.connect(device_id)
        hw.enable_streaming()
        yield hw
        hw.disconnect()
